#!/usr/bin/perl


#    Copyright (C) 2001 Stef Van Dessel
#
#    $Id: bnews-poster.pl.in,v 1.14 2002/06/25 09:57:43 stef Exp $
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


use strict;
use Net::NNTP;
use IO::File;
use Getopt::Long;

my $crlf = "\015\012";

$| = 1;

my $dotdir = (getpwuid($>))[7] . '/.bnews-utils';
my $rcfile = 'postrc';
my $opts = &configure($rcfile);

GetOptions ($opts, 'help!', 'host=s', 'group=s', 'list-queue:s', 'run-queue:s', 'enqueue:s', 'at=s', 'quiet:i', 'from=s', 'username=s', 'password=s', 'subject=s', 'maxpostingsize=s', 'local-hostname=s', 'distribution=s', 'organization=s', 'debug:i', 'verbosity:i', 'escape-dots:i');


$opts->{'help'} and &help;
$opts->{'group'} or die "You must specify a group!\n";
$opts->{'from'} ||= 'nobody@org.invalid';
$opts->{'host'} ||= 'newsbin.telenet.be';
$opts->{'local-hostname'} ||= `hostname`;
$opts->{'bn-encoder'} ||= 'bnews-encoder';
$opts->{'distribution'} ||= 'local';
$opts->{'quiet'} and $opts->{'verbosity'} = 0;

defined $opts->{'bug-compatible'} or $opts->{'bug-compatible'} = 1;
defined $opts->{'escape-dots'} or $opts->{'escape-dots'} = 1;

defined $opts->{'enqueue'} and $opts->{'enqueue'} ||= 'default';
defined $opts->{'run-queue'} and $opts->{'run-queue'} ||= 'default';
defined $opts->{'list-queue'} and $opts->{'list-queue'} ||= 'default';


$opts->{'at'} and &at($opts);
$opts->{'enqueue'} and &enqueue($opts);
$opts->{'list-queue'} and &listqueue($opts);
$opts->{'run-queue'} and &runqueue($opts);

 
defined $opts->{'verbosity'} or $opts->{'verbosity'} = -t STDOUT ? 3 : 2;

$opts->{'maxpostingsize'} ||= 640000;

chomp $opts->{'local-hostname'};

my $maxlines = int($opts->{'maxpostingsize'} / 248);
my $divider = $maxlines * 240;
my $zeroes = " " x 41;

my @topost;
my $file;

$opts->{'group'} =~ s/^a\.b\./alt.binaries./i;

while ($file = shift @ARGV) {
	print "Processing $file\n" if ($opts->{'debug'} > 0);
	if (! -f $file) {
		warn "$file does not exist\n";
		next;
	}
	push(@topost, $file);
}

my $server =  &nntpconnection;

my $fileno = 0;
for $file (@topost) {
	$file =~ m/^(.*\/)?(.*)$/;
	my $lfile = $2;
	$fileno++;
	my $uniqueid = '';
	for (1..10) {
		$uniqueid .= chr(rand(26) + 65);
	}
	my $hdr;
	$hdr  = "From: " . $opts->{'from'} . $crlf;
	$hdr .= "Newsgroups: " . $opts->{'group'} . $crlf;
	$hdr .= "MIME-Version: 1.0" . $crlf;
	$hdr .= "Content-Transfer-Encoding: X-Bommanews-224-40" . $crlf;
	$hdr .= "X-Newsreader: B-news 1.1.4 (Compatible; bnews-utils 20020930)" . $crlf;
	$hdr .= "Content-Type: Application/Octet-Stream; Name=\"" . $lfile . ".bn\"" . $crlf;
	$hdr .= "X-No-Archive: Yes" . $crlf;
	$hdr .= "Distribution: " . $opts->{'distribution'} . $crlf if defined $opts->{'distribution'};
	$hdr .= "Organization: " . $opts->{'organization'} . $crlf if defined $opts->{'organization'};

	my $fsize = (stat $file)[7];
	my $fh = IO::File->new("$opts->{'bn-encoder'} < \"$file\" |") or die "Can't open at pipe ($!)\n";


	my $chunks = int(($fsize - 1) / $divider) + 1 ;

	my $bpst = sprintf("<<BPST File=[%s] Size=%d !BPST>>%s", $lfile, $fsize, $crlf);

	# print "should be posting $chunks chunks..\n";

	for my $chunk (1..$chunks) {

		print "Posting chunk $chunk of $chunks\n" if ($opts->{'verbosity'} >= 2);
		my $hdrsubject = sprintf("Subject: BN %s (%d of %d) -:- %s (%03d/%03d)%s", 
	 		$opts->{'subject'}, $fileno,
			$#topost + 1, $lfile, $chunk, $chunks, $crlf);
		print $hdrsubject if ($opts->{'verbosity'} >= 3 );

		my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
							localtime(time);

		# FIXME: figure out msecs..
 		# use number of chunks for now
		# We need Time::Hires for msecs, but that's not available everywhere
		my $msecs = $chunk;

		my $timedatestamp = sprintf("%02d%02d%02d%02d%02d%02d%03d", ($year - 100), $mon + 1, $mday, $hour, $min, $sec, $msecs);

		my $hdrmessageid = sprintf("Message-ID: <bpst%s.%s@%s>%s", $timedatestamp, $uniqueid, $opts->{'local-hostname'}, $crlf);

		print "$hdrmessageid" if ($opts->{'verbosity'} >= 3);

		
		my $lines = 0;
		# Build the message in core
		#undef(@message);

		my @message;
		push(@message, $hdrsubject);
		push(@message, $hdr);
		push(@message, $hdrmessageid);
		push(@message, $crlf);

		($chunk == 1) && push(@message, $bpst);

		#undef(@out);
		my @out;

		my $line;
		while (($lines++ < $maxlines) and ($line = <$fh>)) {

			# We cannot use $server->datasend since it will do dot-escaping !
			# syswrite is dangerous, FIXME!
			# grrr, bnews-1.1.4 now escapes the dot - this would have made my life
			# easier.


			chomp $line;
			if ($line eq "") {
				print "Skipping empty line\n" if ($opts->{'debug'} > 0);
			} else {
				push(@message, $line . $crlf);
			}
		}
		print "lines = $lines\n" if ($opts->{'verbosity'} >= 3);
		if ($opts->{'bug-compatible'} == 1) {
			if ($chunk == $chunks) {
				my $lastline = pop @message;
				$lastline =~ s/$crlf$//;
		
		#		print "lastline ", length $lastline, " \n" if ($opts->{'debug'} > 0);
				if (((length $lastline) % 41) == 0) {
					if (length $lastline == 246) {
						print "Filling up last line with zeroes\n" if ($opts->{'debug'} > 0);
						push(@message, $lastline . $crlf);
						push(@message, $zeroes . $crlf);
					} else {
						print "Appending zeroes to last line\n" if ($opts->{'debug'} > 0);
						push(@message, $lastline . $zeroes . $crlf);
					}
				} else {
		#				print "Do nothing\n" if ($opts->{'debug'} > 0);
						push(@message, $lastline . $crlf);
				}
			}
		}

		# We cannot use $server->dataend since it will add an extra CR/LF pair
		# before the trailing dot

		my $success = 0;
		do {
			$server->post;
			for my $m (@message) {
				if ($opts->{'escape-dots'}) {
					if (substr($m,0,1) eq '.') {
						substr($m,0,0,'.');
						print "Doing dot escape !\n" if ($opts->{'verbosity'} >= 3);
					}
				}
				syswrite($server, $m);
			}

			syswrite ($server,"." . $crlf);

			my $res = $server->getline;

			chomp $res;
			#$server->debug_print(0, "Got from server: $res");
			
			print "Got back from server: $res\n" if ($opts->{'verbosity'} >= 3);
			if ($res =~ m/^240/) {
				$success = 1
			} else {
				$server->quit;
				$server = &nntpconnection;
			}
		} until $success;
		
		# FIXME
		print "Post Completed, waiting for server to catch up\n" if ($opts->{'verbosity'} >= 3);
		sleep 1;
	}
	$fh->close;

}
$server->quit;

exit 0;

sub configure {
	my $rcfile = shift;
	my $opts = {};
	my $cfg = "$dotdir/$rcfile";

	-f $cfg or return $opts;

	my $fh = IO::File->new($cfg) or die "Can't open config $cfg ($!)\n";

	for my $line ($fh->getlines) {
		$line =~ /^\s*#/ and next;
		$line =~ s/\r//g; chomp $line;

		my ($k, $v) = split(/\s*=\s*/, $line, 2);

		$opts->{$k} = $v;
	}

	$fh->close;

	$opts;
}

sub at {
	my $opts = shift;
	my $at = $opts->{'at'};

	delete $opts->{'at'};

	my $cmd = join(" ", $0, &mkopts($opts));

	my $fh = IO::File->new("|at $at") or die "Can't open at pipe ($!)\n";
	$fh->print ("$cmd\n");
	$fh->close;

	exit 0;
}

sub enqueue {
	my $opts = shift;
	my $q = $opts->{'enqueue'};
	my $qfile = "$dotdir/queues-poster/$q";

	delete $opts->{'enqueue'};

	my $cmd = join(" ", &mkopts($opts));

	mkdir "$dotdir/queues-poster", 0755;

	my $fh = IO::File->new(">>$qfile") or die "Can't open file for queue $q ($!)\n";
	$fh->print ("$cmd\n");
	$fh->close;

	print "QUEUED: $0 $cmd\n";

	exit 0;
}

sub runqueue {
	my $opts = shift;
	my $q = $opts->{'run-queue'};
	my $qfile = "$dotdir/queues-poster/$q";
	my @jobs = ();

	-f $qfile or exit 0;
	while ( -s $qfile ) {
		my $fh = IO::File->new($qfile, "r+") or die "Can't open file for queue $q ($!)\n";
		my @jobs = $fh->getlines;
		my $cmd = pop @jobs;
		$fh->truncate(0);
		$fh->print (@jobs);
		$fh->close;

		print "RUN: $0 $cmd";
		$fh = IO::File->new("|sh") or die "Can't open sh pipe ($!)\n";
		$fh->print ("$0 $cmd");
		$fh->close;
	}

	-z $qfile and unlink "$dotdir/queues/$q";

	exit 0;
}

sub listqueue {
	my $opts = shift;
	my $q = $opts->{'list-queue'};
	my $qfile = "$dotdir/queues-poster/$q";

	-f $qfile or exit 0;

	my $fh = IO::File->new($qfile) or die "Can't open file for queue $q ($!)\n";
	my @jobs = $fh->getlines;
	$fh->close;

	print map { "JOB: $0 $_" } @jobs;

	exit 0;
}

sub mkopts {
	my $opts = shift;
	my @opts = ();

	for my $opt (sort keys %{$opts}) {
		ref($opts->{$opt}) eq 'ARRAY' and push (@opts, join(" ", map { "--$opt='$_'" } @{$opts->{$opt}}));
		ref($opts->{$opt}) eq '' and push (@opts, "--$opt='$opts->{$opt}'");
	}

	@opts;
}

sub help {
	print <<EOT;

Usage: bnews-poster.pl [options]
  
Options:
  
  --host=hostname : news server to use (default=newsbin.telenet.be)
  --group=groupname : group to post to, this is mandatory
  --quiet : be quiet (errors are still printed)
  --help : show this help
  --from=rfc822 : a valid rfc822 address to use in the from header
  --username=user : username to use when authenticating
  --password=mypass : password to use when authenticating
  --subject=subject : post with this subject
  --maxpostingsize : limit messages to this size in bytes
  --local-hostname=blah : local hostname to be used in the message id
  --distribution=local : distribution to use when posting
  --organization=blah : organization to use when posting (default=none)
  --verbosity=n : set verbosity level (default=3 on a tty, 2 elsewhere)
  --escape-dots=n : do we need to escape dots or not (default = 1)
  --enqueue[=queue] : enqueue as a job on specified queue (or default queue)
  --run-queue[=queue] : execute jobs on specified queue (or default queue)
  --list-queue[=queue] : list jobs on specified queue (or default queue)
  --at=timespec : schedule execution using at(1) at given timespec


Examples:

  \$ bnews-poster.pl --host=localhost --group=local.binaries \
	 --subject="test1" *.pl

Default options are read from \$HOME/.bnews-utils/$rcfile if available.


EOT
	exit 0;
}

sub nntpconnection  {
        my $server =  Net::NNTP->new($opts->{'host'},
                 Debug => $opts->{'debug'})
                 or die "Can't connect to news server $opts->{'host'}\n";

        if (defined $opts->{'username'} and defined $opts->{'password'}) {
                $server->command("AUTHINFO USER $opts->{'username'}");
                $server->response;
                $server->command("AUTHINFO PASS $opts->{'password'}");
                $server->response;
        }
        $server;
}


